/*----------------------------------------------------------------------------
 Copyright:
 Author:         Sebastian Voitzsch
 Remarks:
 known Problems: none
 Version:        04.03.2014
 Description:
------------------------------------------------------------------------------*/

#include "main.h"

void hm_dec(unsigned char *data)
{
	uint8_t t, i;

	uint8_t prev = data[1];
	data[1]=(~data[1]) ^ 0x89;		// decode received data
	for (i=2; i < (data[0]); i++)
	{
		t = data[i];
		data[i] = (prev + 0xdc) ^ data[i];
		prev = t;
	}
	data[i] ^= data[2];
}

void hm_enc(unsigned char *data)			// encode message and send via RFM22
{
	uint8_t i;
	data[1] = (~data[1] ^ 0x89);
	uint8_t dat2 = data[2];
	uint8_t prev = data[1];
	for (i = 2; i < data[0]; i++)
	{
		prev = (prev + 0xdc) ^ data[i];
		data[i] = prev;
	}
	data[i] ^= dat2;
}

int main(void)
{
	unsigned char cnt, i, act_status;
	OUT_DDR |= (1 << OUT_P);			// PB7 for switch, PB6 for LED, PB2 SCK, PB3 MOSI)
	OUT_PORT &= ~(1 << OUT_P);			// other ports input and pullup on

	BUTTON_DDR &= ~(1 << BUTTON_P);			// input for button pin
	BUTTON_PORT |= (1 << BUTTON_P);			// pullup for button

	TCCR0B = (1 << CS02 | 1 << CS00);		// CLK/1024 for timer0
	TIMSK0 = (1 << TOIE0);				// enable timer interrupt

	PCMSK0 = (1 << PCINT2 | 1 << PCINT0);		// enable PCINT2 (NIRQ) and PCINT0 (button)
	GIMSK = (1 << PCIE0);				// enable PCINT

	rf22_init();
	act_status = eeprom_read_byte(&eeActStatus);
	if (act_status == 0xff) act_status = 0x00;	// set default value for uninitialized eeprom
	if (act_status == 0xc8)	cbi(OUT_PORT, OUT_P);	// restore former status
	else sbi(OUT_PORT, OUT_P);

	ccu_id[0] = eeprom_read_byte(&eeCCUID1);
	ccu_id[1] = eeprom_read_byte(&eeCCUID2);
	ccu_id[2] = eeprom_read_byte(&eeCCUID3);
	config = 0; counter = 0;
//	if (ccu_id[0] == 0xff)				// eeprom uninitialized?
//	{
//		ccu_id[0]=0x00;
//		ccu_id[1]=0x00;
//		ccu_id[2]=0x00;
//	}

	rf22_rxmode();
	sei();

	while(1)
	{
		if (bit_is_set(BUTTON_PIN, BUTTON_P) && (timer==0) && !(config) && (counter > 2))
		{

			rf22_write(0x06, 0x40);				// enable preamble detect irq
			rf22_write(0x08, 0x04);				// enable low duty cycle mode
			rf22_write(0x07, 0x20);				// enable WUT for rfm22
			MCUCR = (1 << SE | 1 << SM1 );			// main task is...enable sleep mode (INT0 on low level,
			asm volatile ("sleep");				// good bye...

			rf22_read(0x04);				// clear sync word irq status
			rf22_write(0x07, 0x03);				// switch rf22 to rx mode
			rf22_write(0x08, 0x00);				// disable low duty mode
			rf22_write(0x06, 0x00);				// disable preamble detect

			counter = 0;					// reset timer
			while (bit_is_set(NIRQ_PIN, NIRQ_P) && (counter < 1))		// wait until command received after wakeup
				asm volatile ("nop");
		}

		if(cmd == BUTTON_LONG)				// button pressed long?
		{
			cbi(OUT_PORT, OUT_P);			// switch on output to show we're pairing
			hm_msg.length = 0x1a;
			hm_msg.flag=0xA2;
			hm_msg.type=0x00;
			hm_msg.source[0]=HM_ID1;hm_msg.source[1]=HM_ID2;hm_msg.source[2]=HM_ID3;
			memcpy_PF(hm_msg.payload, dev_info, 18);
			hm_enc((unsigned char *)&hm_msg);	// encode homematic string
			rf22_sendpacket((unsigned char *)&hm_msg);
			cmd = 0;
		}
		if (bit_is_clear(NIRQ_PIN, NIRQ_P))		// got message?
		{
			cli();
			cnt = rf22_getpacket((unsigned char *) &hm_msg);
			if (cnt!= 0)
			{
				hm_dec((unsigned char *) &hm_msg);					 // perform homematic decoding
				rec_flag = hm_msg.flag;
				rec_length = hm_msg.length;
				rec_type = hm_msg.type;
				if (hm_msg.destination[0] == HM_ID1 && hm_msg.destination[1] == HM_ID2 && hm_msg.destination[2] == HM_ID3)
				{
					switch (rec_type)
					{
						case 0x01:
							hm_msg.flag = 0x80;				// msg, ack required (prepare ack message)
							hm_msg.type = 0x02;				// channel data
							hm_msg.length = 0x0a;				// set ack length
							switch (hm_msg.payload[1])
							{
								case 0x03:
									if (hm_msg.payload[0] == 1)				// read bank 1 = peers
									{
										hm_msg.length = 0x0e;				// empty for now (type = 0x2 and terminating entry)
										hm_msg.flag = 0xa0;				// ack requested
										hm_msg.type = 0x10;				// message to ccu
										hm_msg.payload[0] = 0x01;
										hm_msg.payload[1] = 0x00;
										hm_msg.payload[2] = 0x00;
										hm_msg.payload[3] = 0x00;
										hm_msg.payload[4] = 0x00;
									}
									break;
								case 0x04:							// read register
									if (hm_msg.payload[0] == 0)				// read bank 0 - config
									{
										hm_msg.length = 0x12;				// send address
										hm_msg.flag = 0xa0;				// ack requested
										hm_msg.type = 0x10;				// message to ccu
										memcpy_PF(hm_msg.payload, reg_bank_0, 6);	// i.e., reg. bank 0,
										hm_msg.payload[2] = ccu_id[0];			// address 0x0a, 0x0b & 0x0c
										hm_msg.payload[4] = ccu_id[1];
										hm_msg.payload[6] = ccu_id[2];
										hm_msg.payload[7] = 0x00;			// terminating entry
										hm_msg.payload[8] = 0x00;
									}
									break;
								case 0x05:							// begin config mode
									config = 1;						// set config flag
									reg_bank = hm_msg.payload[0];
									hm_msg.payload[0]=0x00;					// no data
									break;
								case 0x06:							// end config mode
									config = 0;
									hm_msg.payload[0]=0x00;					// no data
									break;
								case 0x08:							// write reg
									if (hm_msg.payload[0] == 0)				// write config bank
									{
//										if (hm_msg.payload[2] == 2)			// address mode
										{
											for (i=2; i < (rec_length-10); i+=2)
											{
												if (hm_msg.payload[i] == 0x0a)		// fetch address
												{
													ccu_id[0] = hm_msg.payload[i+1];
													eeprom_write_byte(&eeCCUID1, ccu_id[0]);
												}
												if (hm_msg.payload[i] == 0x0b)
												{
													ccu_id[1] = hm_msg.payload[i+1];
													eeprom_write_byte(&eeCCUID2, ccu_id[1]);
												}
												if (hm_msg.payload[i] == 0x0c)
												{
													ccu_id[2] = hm_msg.payload[i+1];
													sbi(OUT_PORT, OUT_P);			// valid address received, switch off pairing LED
													eeprom_write_byte(&eeCCUID3, ccu_id[2]);
												}
											}
										}
									}
									break;

								case 0x0e:
									hm_msg.length = 0x0e;
									hm_msg.type = 0x10;
									hm_msg.flag = 0xa4;
									hm_msg.payload[0] = 0x06;	// actuator status
									hm_msg.payload[1] = 0x01;	// channel 1
									hm_msg.payload[2] = act_status;	// state
									hm_msg.payload[3] = 0x00;	// rssi
									hm_msg.payload[4] = 0x00;	// battery good
							}
							break;

						case 0x02:			// ack
							hm_msg.flag = 0x00;	// clear flag to avoid sending ack on ack...
							hm_msg.length = 0;
							break;

						case 0x11:			// command: set state
							act_status = hm_msg.payload[2];				// save state
							eeprom_write_byte(&eeActStatus, act_status);		// write to eeprom
							if (act_status == 0xc8)
							{
								cbi(OUT_PORT, OUT_P);
							}
							else if (act_status == 0x00)
							{
								sbi(OUT_PORT, OUT_P);
							}
							hm_msg.payload[0] = 0x01;
							hm_msg.type = 0x02;
							hm_msg.flag = 0x80;
							break;
					}

					if (rec_flag & 0x80)		// ack needed?
					{
						hm_msg.destination[0]=hm_msg.source[0];hm_msg.destination[1]=hm_msg.source[1];hm_msg.destination[2]=hm_msg.source[2];
						hm_msg.source[0]=HM_ID1;hm_msg.source[1]=HM_ID2;hm_msg.source[2]=HM_ID3;
						hm_enc((unsigned char *)&hm_msg);		// encode message
						rf22_sendpacket((unsigned char *)&hm_msg);	// send it
						hm_dec((unsigned char *)&hm_msg);
					}
				}
				hm_msg.destination[0] = 0;		// clear destination
				cnt = 0;
			}
			rf22_rxmode();					// enable rx again and clear irq status
			sei();
		}
	}
}

ISR (TIM0_OVF_vect)
{
	counter++;
	if (bit_is_clear(BUTTON_PIN, BUTTON_P))			// button pressed?
	{
		timer++;
		if(timer > 11)
		{
			cmd = BUTTON_LONG;
			timer = 0;
		}
	}
	if (bit_is_set(BUTTON_PIN, BUTTON_P) && timer)		// button release?
	{
		if (timer < 3)
		{
			cmd = BUTTON_SHORT;
		}
		timer = 0;
	}
}

ISR(PCINT0_vect)
{
//	asm volatile("nop");					// stub, only for wakeup from IRQ
}
