/*----------------------------------------------------------------------------
 Copyright:
 Author:
 Remarks:
 known Problems: none
 Version:        16.12.2009
 Description:    RF22 Funkmodule
------------------------------------------------------------------------------*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "rf22.h"
#include "usart.h"
#include "rf22conf.h"

// transfer in burst mode (i.e., CS already low

static inline unsigned char rf22_burst(unsigned char data)
{

	USIDR = data;
	for (unsigned char i=0; i<16; i++)
	{
		USICR = (1 << USIWM0 | 1 << USICS1 | 1 << USICLK | 1 << USITC);	// generate clk
//		_delay_us(1);
	}
	return USIDR;
}

// begin burst transmission (r/w selected with bit 8)

static inline unsigned char rf22_start_burst(unsigned char addr)
{
	RF22_PORT &= ~(1 << CS);
//	_delay_us(3);
	return rf22_burst(addr);
}

// end burst mode

void rf22_end_burst(void)
{
	RF22_PORT |= (1 << CS);
}

// spi transmission function

static inline unsigned char rf22_trans(unsigned char addr, unsigned char data)
{
	unsigned char res;

//	rf22_start_burst(addr);
//	res = rf22_burst(data);
//	rf22_end_burst;
	RF22_PORT &= ~(1 << CS);		// chip select enable
//	_delay_us(3);
	rf22_burst(addr);
	res = rf22_burst(data);
	RF22_PORT |= (1 << CS);
	return res;		// return result
}

//###############################################################################################################
void rf22_write(unsigned char addr, unsigned char data)
{
	rf22_trans(128|addr,data);
}


//###############################################################################################################
unsigned char rf22_read(unsigned char addr)
{
	return rf22_trans(addr&127,0);
}


//###############################################################################################################
void rf22_init(void)
{
	RF22_DDR |= (1 << SCK | 1 << DO | 1 << CS );	// initialize RF22 port for SPI
	RF22_PORT |= (1 << DI | 1 << CS | 1 << NIRQ_P);				// pullup on DI/NIRQ pin, CS high
//	NIRQ_DDR &= ~(1 << NIRQ_P);
//	NIRQ_PORT |= (1 << NIRQ_P);

	_delay_ms(10); 	// wait until POR done
	rf22_write(0x07, 0x80);					// software reset
	_delay_ms(10); 	// wait until software reset done

	for (unsigned char i=0; i < sizeof(conf_array)+1; i+=2)
	{
		rf22_write(pgm_read_byte(&conf_array[i]),pgm_read_byte(&conf_array[i+1]));	// write config data 
	}
}

//###############################################################################################################
void rf22_rxmode(void)
{
	rf22_write(0x07, 0x01);		// to_ready_mode();
	rf22_write(0x7e, 0x17);		// threshold for rx almost full, interrupt when 1 byte received
	rf22_write(0x08, 0x07);		// clear RX fifo
	rf22_write(0x08, 0x00);		// clear fifo, disable multi packet, enable low duty cycle
	rf22_write(0x07, 0x05);		// RX on

	rf22_read(0x03);		// clear interrupt status
	rf22_read(0x04);		// clear interrupt status
}


//###############################################################################################################
void rf22_sendpacket(unsigned char *data)
{
	unsigned char i, size;
	uint16_t crc;
        uint8_t carry = 1;
        uint8_t xor_key = 0xff;

	size = data[0];
	if (size>40) size=40;		// max size - crc (2 bytes) and length byte

	crc = calc_CRC(data, size);
	data[size+1] = (crc >> 8);		// save high and
	data[size+2] = crc & 0xff;	// low crc byte

	rf22_write(0x07, 0x03);		// switch to ready mode
	rf22_write(0x3e, size+3);	// packet length

	rf22_write(0x08, 0x01);		// clear TX fifo
	rf22_write(0x08, 0x04);		// clear TX fifo

	rf22_read(0x03);		// clear interrupt status
	rf22_read(0x04);		// clear interrupt status

	rf22_start_burst(0xff);		// write burst to 0x7f
	for (i = 0; i<size+3; i++)	// data + length + 2 crc bytes
	{
		rf22_burst((data[i] ^ xor_key));	// write whitened data
		get_next_key(&xor_key, &carry);
	}

	rf22_end_burst();
	rf22_write(0x07, 0x09);		// TX on

	while (bit_is_set(NIRQ_PIN, NIRQ_P));

	rf22_write(0x3e, 0x40);		// restore max. packet length
	rf22_rxmode();
}


//###############################################################################################################
unsigned char rf22_getpacket(unsigned char *data)
{
	unsigned char rx_buf[MAX_PKT_SIZE+2];
	unsigned char i, size, tmp;
	uint8_t carry = 1;
	uint8_t xor_key = 0xff;
	uint16_t crc;

//	if(rf22_read(0x31)&0x1A)		// receiving a packet
//	if(rf22_read(0x31)&0x08)
	{
		if ((rf22_read(0x03)&2)&&(!(rf22_read(0x02)&32)))	// packet received & not read
		{
			size=rf22_read(0x7f) ^ xor_key;	// packet length (dewhitened)
			data[0]=size;			// save length
			rf22_start_burst(0x7f);
			for (i=0; i<size+2; i++)		// Daten (cnt + 2 to get CRC as well)
			{
				get_next_key(&xor_key, &carry);		// get dewhitening key
				data[i+1]=rf22_burst(0) ^ xor_key;	// read and dewhite data
			}

			rf22_end_burst();
//			rf22_rxmode();
			crc = calc_CRC(data, size);
			data[i+1]=crc >> 8;
			data[i+2]=crc & 0xff;
			return (size);
		}
	}
	return 0;
}

// do data whitening in software
// as RFM22 does it different from CC110x
// initialize with key = 0xff and carry = 0b1

void get_next_key(uint8_t *key, uint8_t *carry)
{

	uint8_t bit_five, bit_zero, carry_new, i;

	for (i=0;i < 8; i++)
	{
		bit_five = (*key & (1 << 5)) >> 5;
		bit_zero = (*key & (1 << 0)) >> 0;
		carry_new = bit_five ^ bit_zero;
		*key >>= 1;
		*key |= (*carry << 7);
		*carry = carry_new;
	}
}

// routine to do the crc calculation as RFM22 has different CRC mechanism
// according to TI appnote DN502

uint16_t calc_CRC (uint8_t *data, uint8_t length)
{
uint8_t i, i1, bte;
uint16_t checksum = 0xffff;

	for (i=0; i < length+1; i++)		// include length byte in CRC calc
	{
		bte = data[i];
		for (i1=0; i1 < 8; i1++)
		{
			if (((checksum & 0x8000) >> 8) ^ (bte & 0x80))
				checksum = (checksum << 1) ^ 0x8005;
			else
				checksum = (checksum << 1);
			bte <<= 1;
		}
	}
	return checksum;
}
